Neko 1.99.9
A portable framework for high-order spectral element flow simulations
Loading...
Searching...
No Matches
opr_cdtp.hip
Go to the documentation of this file.
1/*
2 Copyright (c) 2021-2026, The Neko Authors
3 All rights reserved.
4
5 Redistribution and use in source and binary forms, with or without
6 modification, are permitted provided that the following conditions
7 are met:
8
9 * Redistributions of source code must retain the above copyright
10 notice, this list of conditions and the following disclaimer.
11
12 * Redistributions in binary form must reproduce the above
13 copyright notice, this list of conditions and the following
14 disclaimer in the documentation and/or other materials provided
15 with the distribution.
16
17 * Neither the name of the authors nor the names of its
18 contributors may be used to endorse or promote products derived
19 from this software without specific prior written permission.
20
21 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
22 "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
23 LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
24 FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
25 COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
26 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
27 BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
28 LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
29 CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
30 LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
31 ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
32 POSSIBILITY OF SUCH DAMAGE.
33*/
34
35#include <string.h>
36#include <stdlib.h>
37#include <stdio.h>
38#include <hip/hip_runtime.h>
40#include <device/hip/check.h>
41#include "cdtp_kernel.h"
42#include "elem_block_tune.h"
43
44extern "C" {
45 #include <common/neko_log.h>
46}
47
48template < const int >
49int tune_cdtp(void *dtx, void *x,
50 void *dr, void *ds, void *dt,
51 void *dxt, void *dyt, void *dzt,
52 void *w3, int *nel, int *lx, int *eb_sel, int *ch_sel,
53 int *nwf_sel);
54
55extern "C" {
56
60 void hip_cdtp(void *dtx, void *x,
61 void *dr, void *ds, void *dt,
62 void *dxt, void *dyt, void *dzt,
63 void *w3, int *nel, int *lx) {
64
65 static int autotune[17] = { 0 };
66 /* elements per block candidate chosen by the tuner */
67 static int autotune_eb[17] = { 0 };
68 /* chunk candidate chosen for the 1d variant */
69 static int autotune_ch[17] = { 0 };
70 /* wavefronts per block candidate chosen for the mfma variant */
71 static int autotune_nwf[17] = { 0 };
72
73 const dim3 nthrds_1d(1024, 1, 1);
74 const dim3 nthrds_kstep((*lx), (*lx), 1);
75 const dim3 nblcks((*nel), 1, 1);
76
77#define CASE_1D(LX, C) \
78 hipLaunchKernelGGL( HIP_KERNEL_NAME( \
79 cdtp_kernel_1d<real, LX, NEKO_CHUNKS(LX, C)> ), \
80 nblcks, NEKO_CHUNKS_NTHRDS(LX, C), 0, \
81 (hipStream_t) glb_cmd_queue, \
82 (real *) dtx, (real *) x, \
83 (real *) dr, (real *) ds, (real *) dt, \
84 (real *) dxt, (real *) dyt, (real *) dzt, \
85 (real *) w3); \
86 HIP_CHECK(hipGetLastError());
87
88/* Runtime dispatch onto the tuned chunk candidate */
89#define CASE_1D_SEL(LX, SEL) \
90 switch (SEL) { \
91 case 0: CASE_1D(LX, 0); break; \
92 case 1: CASE_1D(LX, 1); break; \
93 case 2: CASE_1D(LX, 2); break; \
94 default: CASE_1D(LX, 3); break; \
95 }
96
97#define CASE_KSTEP(LX, C) \
98 hipLaunchKernelGGL( HIP_KERNEL_NAME( \
99 cdtp_kernel_kstep<real, LX, NEKO_EB(LX, C)> ), \
100 NEKO_EB_NBLCKS(*nel, LX, C), NEKO_EB_NTHRDS(LX, C), 0, \
101 (hipStream_t) glb_cmd_queue, \
102 (real *) dtx, (real *) x, \
103 (real *) dr, (real *) ds, (real *) dt, \
104 (real *) dxt, (real *) dyt, (real *) dzt, \
105 (real *) w3, *nel); \
106 HIP_CHECK(hipGetLastError());
107
108/* Runtime dispatch onto the tuned candidate */
109#define CASE_KSTEP_SEL(LX, SEL) \
110 switch (SEL) { \
111 case 0: CASE_KSTEP(LX, 0); break; \
112 case 1: CASE_KSTEP(LX, 1); break; \
113 default: CASE_KSTEP(LX, 2); break; \
114 }
115
116#define CASE_MFMA(LX, C) \
117 hipLaunchKernelGGL( HIP_KERNEL_NAME( \
118 cdtp_kernel_mfma<real, LX, NEKO_MFMA_NWF(C), \
119 NEKO_MFMA_TILE(C)> ), \
120 NEKO_MFMA_NBLCKS(*nel, LX, C), NEKO_MFMA_NTHRDS(C), 0, \
121 (hipStream_t) glb_cmd_queue, \
122 (real *) dtx, (real *) x, \
123 (real *) dr, (real *) ds, (real *) dt, \
124 (real *) dxt, (real *) dyt, (real *) dzt, \
125 (real *) w3, *nel); \
126 HIP_CHECK(hipGetLastError());
127
128/* Runtime dispatch onto the tuned candidate, which names both the wavefronts
129 per block and the matrix core tile, see NEKO_MFMA_TILE in mfma_kernel.h */
130#define CASE_MFMA_SEL(LX, SEL) \
131 switch (SEL) { \
132 case 0: CASE_MFMA(LX, 0); break; \
133 case 1: CASE_MFMA(LX, 1); break; \
134 case 2: CASE_MFMA(LX, 2); break; \
135 case 3: CASE_MFMA(LX, 3); break; \
136 case 4: CASE_MFMA(LX, 4); break; \
137 case 5: CASE_MFMA(LX, 5); break; \
138 case 6: CASE_MFMA(LX, 6); break; \
139 case 7: CASE_MFMA(LX, 7); break; \
140 case 8: CASE_MFMA(LX, 8); break; \
141 default: CASE_MFMA(LX, 9); break; \
142 }
143
144/* The switch above has to name every candidate, see the note in ax_helm.hip:
145 a candidate reaching the default arm launches one geometry's grid at
146 another geometry's kernel. */
147static_assert(NEKO_MFMA_CANDIDATES == 10,
148 "CASE_MFMA_SEL does not cover the candidate space");
149
150#define CASE(LX) \
151 case LX: \
152 if(autotune[LX] == 0 ) { \
153 autotune[LX]=tune_cdtp<LX>(dtx, x, \
154 dr, ds, dt, \
155 dxt, dyt, dzt, \
156 w3, nel, lx, &autotune_eb[LX], \
157 &autotune_ch[LX], \
158 &autotune_nwf[LX]); \
159 } else if (autotune[LX] == 1 ) { \
160 CASE_1D_SEL(LX, autotune_ch[LX]); \
161 } else if (autotune[LX] == 2 ) { \
162 CASE_KSTEP_SEL(LX, autotune_eb[LX]); \
163 } else if (autotune[LX] == 3 ) { \
164 CASE_MFMA_SEL(LX, autotune_nwf[LX]); \
165 } \
166 break
167
168#define CASE_LARGE(LX) \
169 case LX: \
170 CASE_KSTEP(LX, 0); \
171 break
172
173
174 if ((*lx) < 13) {
175 switch(*lx) {
176 CASE(2);
177 CASE(3);
178 CASE(4);
179 CASE(5);
180 CASE(6);
181 CASE(7);
182 CASE(8);
183 CASE(9);
184 CASE(10);
185 CASE(11);
186 CASE(12);
187 default:
188 {
189 fprintf(stderr, __FILE__ ": size not supported: %d\n", *lx);
190 exit(1);
191 }
192 }
193 }
194 else {
195 switch(*lx) {
196 CASE_LARGE(13);
197 CASE_LARGE(14);
198 CASE_LARGE(15);
199 CASE_LARGE(16);
200 default:
201 {
202 fprintf(stderr, __FILE__ ": size not supported: %d\n", *lx);
203 exit(1);
204 }
205 }
206 }
207 }
208}
209
210template < const int LX >
211int tune_cdtp(void *dtx, void *x,
212 void *dr, void *ds, void *dt,
213 void *dxt, void *dyt, void *dzt,
214 void *w3, int *nel, int *lx, int *eb_sel, int *ch_sel,
215 int *nwf_sel) {
218 int best1 = 0;
221 int best3 = 0;
222 const int rounds = neko_tune_rounds();
223 const int iters = neko_tune_iters();
224 /* Candidates of the kstep sweep, one -- the unblocked shape -- when the
225 elements per block sweep is off */
226 const int eb_cand = neko_eb_sweep() ? NEKO_EB_CANDIDATES : 1;
227 /* Geometry pinned by each formulation's own variable, -1 to sweep it */
228 const int ch_pin = neko_chunks_pin();
229 const int eb_pin = neko_eb_pin();
230 /* Candidates of the matrix core sweep, see neko_mfma_candidates() */
231 const int mfma_cand = neko_mfma_candidates();
232 /* Both matrix core dimensions, pinned by their own variables or -1 */
233 const int nwf_pin = neko_mfma_pin();
234 /* Formulation pinned by NEKO_AUTOTUNE, as the identifier this returns */
235 int strat = 0;
236 const bool mfma = mfma_lx_supported<LX>() && hip_have_mfma();
237 int best = 0;
238 int retval;
239
240 for (int c = 0; c < NEKO_EB_CANDIDATES; c++) {
242 }
243 for (int c = 0; c < NEKO_CHUNKS_CANDIDATES; c++) {
245 }
246 for (int c = 0; c < NEKO_MFMA_CANDIDATES; c++) {
248 }
249
250 const dim3 nthrds_1d(1024, 1, 1);
251 const dim3 nthrds_kstep((*lx), (*lx), 1);
252 const dim3 nblcks((*nel), 1, 1);
253 const hipStream_t stream = (hipStream_t) glb_cmd_queue;
254
255 char *env_value = NULL;
256 char neko_log_buf[80];
257
258 env_value=getenv("NEKO_AUTOTUNE");
259
260 sprintf(neko_log_buf, "Autotune cdtp (lx: %d)", *lx);
262
263 *eb_sel = 0;
264 *ch_sel = 0;
265 *nwf_sel = 0;
266
267 /*
268 * NEKO_AUTOTUNE names a formulation and nothing more: the sweep below is
269 * narrowed to that kernel family, but its geometry -- the chunk size, the
270 * elements per block, the wavefronts per block -- is still measured
271 * candidate against candidate
272 */
273 if(env_value) {
274 if( !strcmp(env_value,"1D") ) {
275 strat = 1;
276 } else if( !strcmp(env_value,"KSTEP") ) {
277 strat = 2;
278 } else if( !strcmp(env_value,"MFMA") ) {
279 if (mfma) {
280 strat = 3;
281 } else {
282 sprintf(neko_log_buf, "MFMA strategy not available for this config");
284 }
285 } else {
286 sprintf(neko_log_buf, "Invalid value set for NEKO_AUTOTUNE");
288 }
289 }
290
291 /* Geometry of the pinned formulation, if its own variable fixes that too.
292 Both pinned leaves nothing to measure, so the kernel is launched once and
293 reported, which is what pinning has always done */
294 const int pin = (strat == 1) ? ch_pin : (strat == 2) ? eb_pin :
295 (strat == 3) ? nwf_pin : -1;
296
297 if (pin >= 0) {
298 switch (strat) {
299 case 1:
300 *ch_sel = pin;
302 sprintf(neko_log_buf, "Set by env : 1 (1D, %d chunk)",
304 break;
305 case 2:
306 *eb_sel = pin;
308 sprintf(neko_log_buf, "Set by env : 2 (KSTEP, %d elem/block)",
309 NEKO_EB_SEL(LX, pin));
310 break;
311 default:
312 *nwf_sel = pin;
314 sprintf(neko_log_buf, "Set by env : 3 (MFMA %s, %d wf, %d elem/blk)",
316 break;
317 }
320 return strat;
321 }
322
323 if (strat) {
324 sprintf(neko_log_buf, "Set by env : %d (%s)", strat, env_value);
326 }
327
328 /* Formulations the sweep considers, see NEKO_TUNE_FOR(). NEKO_MFMA_TUNE
329 keeps the matrix core variant out of an unpinned sweep, but naming it
330 explicitly still measures it */
331 const bool try_1d = (strat == 0 || strat == 1);
332 const bool try_kstep = (strat == 0 || strat == 2);
333 const bool try_mfma = mfma && ((strat == 3) ||
334 (strat == 0 && neko_mfma_sweep()));
335
338
339 /* Verify every candidate against the reference formulation on a synthetic
340 input before anything is timed, see neko_tune_verify_begin() in
341 elem_block_tune.h. A candidate that does not reproduce it is reported
342 and dropped from the comparison below. */
343 neko_tune_rej_t rej1 = {{0}}, rej2 = {{0}}, rej3 = {{0}};
344 {
346 real *vouts[] = { (real *) dtx };
347 void **vins[] = { &x };
348 if (neko_tune_verify_begin(&vf, vouts, 1, vins, 1,
349 (*nel) * LX * LX * LX, stream)) {
350 sprintf(neko_log_buf, "Verify : rel tol %.1e, synthetic input",
351 vf.tol);
353 CASE_1D_SEL(LX, 0);
354 neko_tune_verify_ref(&vf, stream);
357 }
360 }
363 }
364 neko_tune_verify_end(&vf, stream);
365 } else {
366 sprintf(neko_log_buf, "Verify : %s", neko_tune_verify_on() ?
367 "skipped, no memory for it" : "off (NEKO_TUNE_VERIFY=0)");
369 }
370 }
371
372 /* Warm every variant before timing anything: each specialisation has to be
373 resident and the clocks at steady state, or whichever is timed first is
374 measured on a colder part */
375 for (int i = 0; i < NEKO_TUNE_WARMUP; i++) {
377 CASE_1D_SEL(LX, c);
378 }
380 CASE_KSTEP_SEL(LX, c);
381 }
383 CASE_MFMA_SEL(LX, c);
384 }
385 }
386
387 /* Interleaved rounds, best time per variant */
388 for (int r = 0; r < rounds; r++) {
391 }
394 }
397 }
398 }
399
400 /* Candidates the verification pass rejected drop out here, after the
401 timing, so the sweep's own bookkeeping is untouched */
405
408
412 *eb_sel = best;
413 *ch_sel = best1;
414 *nwf_sel = best3;
415
416 if (time1[best1] < time2[best]) {
417 retval = 1;
418 } else {
419 retval = 2;
420 }
421
422 /* The mfma variant joins the comparison only where it exists, its
423 candidates are left at NEKO_TUNE_INIT otherwise */
424 if (time3[best3] < ((retval == 1) ? time1[best1] : time2[best])) {
425 retval = 3;
426 }
427
428 /* Leave the chosen kernel's output in place: the tuner stands in for a real
429 evaluation and the variants do not sum in the same order */
430 if (retval == 1) {
432 } else if (retval == 2) {
434 } else {
436 }
437
438 if (retval == 1) {
439 sprintf(neko_log_buf, "Chose : 1 (1D, %d chunk)",
441 } else if (retval == 2) {
442 sprintf(neko_log_buf, "Chose : 2 (KSTEP, %d elem/block)",
444 } else {
445 sprintf(neko_log_buf, "Chose : 3 (MFMA %s, %d wf, %d elem/blk)",
448 }
451 return retval;
452}
__global__ void ale_add_kinematics_kernel(const int n, T *__restrict__ wx, T *__restrict__ wy, T *__restrict__ wz, const T *__restrict__ x_ref, const T *__restrict__ y_ref, const T *__restrict__ z_ref, const T *__restrict__ phi, const T *__restrict__ x, const T *__restrict__ y, const T *__restrict__ z, const kinematics_params_t kin_params)
const int i
__global__ void const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ dyt
__global__ void const T *__restrict__ const T *__restrict__ const T *__restrict__ ds
__global__ void const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ dzt
__global__ void const T *__restrict__ x
__global__ void const T *__restrict__ const T *__restrict__ dr
__global__ void const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ dt
__global__ void const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ dxt
__global__ void const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ w3
#define NEKO_CHUNKS_CANDIDATES
Definition elem_block.h:125
#define NEKO_EB_CANDIDATES
Definition elem_block.h:63
#define NEKO_EB_SEL(LX, SEL)
Definition elem_block.h:108
#define NEKO_CHUNKS_SEL(LX, SEL)
Definition elem_block.h:147
static int neko_chunks_pin()
#define NEKO_TUNE_TIME(T, LAUNCH, LX, C, ITERS)
static int neko_tune_rounds()
static int neko_eb_pin()
#define NEKO_TUNE_LOG(LX, T1, T2)
#define NEKO_TUNE_INIT
#define NEKO_TUNE_BEST(T, BEST, N)
static int neko_tune_iters()
#define NEKO_TUNE_FOR(C, ON, PIN, N)
static int neko_eb_sweep()
#define NEKO_TUNE_WARMUP
double real
#define HIP_CHECK(err)
Definition check.h:8
static void neko_tune_verify_ref(neko_tune_verify_t *v, hipStream_t stream)
#define NEKO_TUNE_VERIFY_CAND(VF, REJ, LAUNCH, LX, C, NAME)
static void neko_tune_verify_end(neko_tune_verify_t *v, hipStream_t stream)
static int neko_tune_verify_on()
static int neko_tune_verify_begin(neko_tune_verify_t *v, real **outs, const int nout, void ***ins, const int nin, const int n, hipStream_t stream)
#define NEKO_TUNE_APPLY_REJ(T, REJ, N)
#define NEKO_TUNE_LOG_MFMA(LX, T3)
static const char * mfma_tile_name(const int c)
#define NEKO_MFMA_CANDIDATES
static int neko_mfma_pin()
static int neko_mfma_candidates()
static bool hip_have_mfma()
#define NEKO_MFMA_NWF(C)
#define NEKO_MFMA_EB(LX, C)
static int neko_mfma_sweep()
void log_error(char *msg)
void log_message(char *msg)
void log_end_section()
void log_section(char *msg)
int tune_cdtp(void *dtx, void *x, void *dr, void *ds, void *dt, void *dxt, void *dyt, void *dzt, void *w3, int *nel, int *lx, int *eb_sel, int *ch_sel, int *nwf_sel)
Definition opr_cdtp.hip:211
#define CASE_KSTEP_SEL(LX, SEL)
#define CASE(LX)
#define CASE_1D_SEL(LX, SEL)
void hip_cdtp(void *dtx, void *x, void *dr, void *ds, void *dt, void *dxt, void *dyt, void *dzt, void *w3, int *nel, int *lx)
Definition opr_cdtp.hip:60
#define CASE_MFMA_SEL(LX, SEL)
#define CASE_LARGE(LX)